//
//  main.m
//  iphoneViewer
//
//  Created by Thomas Hogarth on 10/05/2009.
//  Copyright HogBox 2009. All rights reserved.
//

#import <UIKit/UIKit.h>

int main(int argc, char *argv[]) {
    
    NSAutoreleasePool *pool = [NSAutoreleasePool new];
    UIApplicationMain(argc, argv, nil, @"iphoneViewerAppDelegate");
    [pool release];
    return 0;
}
