rem mew 2004-01-21
rem for description of the icon creation process, see:
rem http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnwxp/html/winxpicons.asp

png2ico.exe osg.ico osg48-32.png osg32-32.png osg16-32.png osg48-8.png osg32-8.png osg16-8.png osg48-4.png osg32-4.png osg16-4.png

