echo osgversion
osgversion
 
echo osgunittests
osgunittests matrix
osgunittests qt
osgunittests sizeof

echo osglauncher
osglauncher

echo osganimate
osganimate

echo osgautotransform
osgautotransform

echo osgbillboard
osgbillboard

echo osgblendequation
osgblendequation

echo osgcallback
osgcallback

echo osgcatch
# osgcatch

echo osgclip
osgclip

echo osgcubemap
osgcubemap

echo osgdistortion
osgdistortion

echo osgforest
osgforest

echo osgfxbrowser
osgfxbrowser

echo osgoutline cow.osgt
osgoutline cow.osgt

echo osggeometry
osggeometry

echo osghangglide
osghangglide

echo osghud
osghud dumptruck.osgt

echo osgimpostor
osgimpostor

echo osgkeyboard
osgkeyboard

echo osgkeyboardmouse
osgkeyboardmouse dumptruck.osgt

echo osglight
osglight glider.osgt

echo osglightpoint
osglightpoint

echo osglogo
osglogo

echo osglogicop
osglogicop glider.osgt

echo osgmotionblur
osgmotionblur cow.osgt

echo osgmultitexture
osgmultitexture cessnafire.osgt

echo osgoccluder
osgoccluder glider.osgt

echo osgocclusionquery
osgocclusionquery

echo osgparticle
osgparticle

echo osgparticleeffects
osgparticleeffects

echo osgpick
osgpick fountain.osgt

echo osgpointsprite
osgpointsprite

echo osgplanets
osgplanets

echo osgprecipitation --rain 0.5
osgprecipitation --rain 0.5 lz.osgt

echo osgprecipitation --snow 0.5
osgprecipitation --snow 0.5 lz.osgt

echo osgprerender
osgprerender dumptruck.osgt

echo osgprerendercubemap
osgprerendercubemap

echo osgreflect
osgreflect cessna.osgt

echo osgscalarbar
osgscalarbar

echo osgscribe
osgscribe cow.osgt

echo osgsequence
osgsequence cow.osgt dumptruck.osgt cessna.osgt glider.osgt

echo osgshaders
osgshaders

echo osgshaderterrain
osgshaderterrain

echo osgparametric
osgparametric

echo osgthirdpersonview
osgthirdpersonview cow.osgt

echo osgviewer glsl_simple.osgt
osgviewer glsl_simple.osgt

echo osgviewer glsl_confetti.osgt
osgviewer glsl_confetti.osgt

echo osgviewer glsl_mandelbrot.osgt
osgviewer glsl_mandelbrot.osgt

echo osgviewer glsl_julia.osgt
osgviewer glsl_julia.osgt

echo osgshape
osgshape

echo osgshadow -3 --sm
osgshadow -3 --sm

echo osgshadow -2 --st
osgshadow -2 --st

echo osgshadow -3 --ssm
osgshadow -3 --ssm

echo osgsimplifier
osgsimplifier --ratio 0.1 cessna.osgt

echo osgsimulation
osgsimulation

echo osgspheresegment
osgspheresegment

echo osgspotlight
osgspotlight

echo osgstereoimage
osgstereoimage

echo osgteapot
osgteapot

echo osgtessellate
osgtessellate

echo osgtext
osgtext

echo osgtexture1D
osgtexture1D

echo osgtexture2D
osgtexture2D

echo osgtexture3D
osgtexture3D

echo osgtexturerectangle
osgtexturerectangle

echo osgvertexprogram
osgvertexprogram

echo osgwindows
osgwindows
